/******************************************************************************
*
* project name: TI-Con45
* initial date: 27/06/2000
* authors:      thomas.nussbaumer@gmx.net (coding)
                marcos.lopez@wol.es (graphics/minor improvements/BETA testing)
* description:  connect-4/connect-5 game for TI-89/92+
*
*
* This program is based on source code of Keith Pomakis (pomakis@pobox.com),
* which was ported and slightly modified to run on a TI89/92+ by me.
*
* Due to the restrictions of the calculator not all possible computer levels
* are used and the size of the board is restricted also to a maximum of
* 12x7 squares.
*
* NOTE: the calculator may need much time to calculate its move on large boards
*
*******************************************************************************/

#define SAVE_SCREEN        // restore screen after program end
#define NO_EXIT_SUPPORT
//#define OPTIMIZE_ROM_CALLS
#define USE_FLINE_ROM_CALLS
#define USE_INTERNAL_FLINE_EMULATOR

#include <tigcclib.h>

#define PRG_NAME       "TI-Con45"
#define PRG_VERSION    "1.41"
#define TITLE_AND_VERSION PRG_NAME" v"PRG_VERSION
#define PRG_SUBVERSION ""

/*===========================================================================*/
/* "waits" a given number of milliseconds (dummy looping)                    */
/* NOTE: this wait loop is parametrized to fit on HW1 calculators            */
/*===========================================================================*/

void WaitForMillis(unsigned short);
asm("WaitForMillis:"\
"       movem.l %d0-%d3,-(%sp);"\
"       move.w  (20,%sp),%d2;"\
"       move.l  #31,%d1;"\
"       move.l  #31,%d3;"\
"_wl2_: move.w  #120,%d0;"\
"_wl1_: rol.l   %d3,%d1;"\
"       dbra    %d0,_wl1_;"\
"       dbra    %d2,_wl2_;"\
"       movem.l (%sp)+,%d0-%d3;"\
"       rts");


enum {HUMAN = 0, COMPUTER = 1};  // type of player


/*---------------------------------------------------------------------------*/
/* global variables                                                          */
/*---------------------------------------------------------------------------*/

SCR_RECT       fullscreen  = {{0, 0, 239, 127}}; // used for fullscreen clipping
short          m_width     = 7;    // number of columns
short          m_height    = 6;    // number of lines
short          m_offsetx   = 0;    // offset due to the actual board size
short          m_offsety   = 0;    // offset due to the actual board size
short          m_think_l   = 0;    // computer thinking message length
short          m_think_x   = 0;    // computer thinking x position
short          m_think_y   = 0;    // computer thinking y position
char*          m_think_m   = (char*)"Thinking ...";  // computer thinking message

unsigned char* plane1      = 0;    // start address of grayscale bitplane 1
unsigned char* plane2      = 0;    // start address of grayscale bitplane 2

#define ti92_offsetx C89_92(0,40)
#define ti92_offsety C89_92(0,14)
//short          ti92_offsetx  = 0; // additional offset for TI92+
//short          ti92_offsety  = 0; // additional offset for TI92+
short          actual_column = 0; // location of the "cursor"


/*---------------------------------------------------------------------------*/
/* sprite data                                                               */
/*---------------------------------------------------------------------------*/

// cleaner sprite - just used for masking
static unsigned short cleaner[] = {
    0x000F,0x000F,0x000F,0x000F,0x000F,0x000F,0x000F,0x000F,
    0x000F,0x000F,0x000F,0x000F,0x000F
};

// white piece - plane 1
static unsigned short white1[] = {
    0x1080,0x5fa0,0x36c0,0xf1f0,0x40a0,0x6060,0x6060,0x4020,
    0xf0f0,0x36c0,0x5fa0,0x1080
};

// white piece - plane 2
static unsigned short white2[] = {
    0x0f00,0x3fc0,0x79e0,0x6360,0xe0f0,0xc0b0,0xc030,0xe070,
    0x6060,0x79e0,0x3fc0,0x0f00
};

// black piece - plane 1
static unsigned short black1[] = {
    0x30c0,0x5fa0,0xbfd0,0xfef0,0x7f60,0x7fe0,0x7fe0,0x7fe0,
    0xfff0,0xbfd0,0x5fa0,0x30c0
};

// black piece - plane 2
static unsigned short black2[] = {
    0x0f00,0x3fc0,0x7fe0,0x7ce0,0xff70,0xff70,0xfff0,0xfff0,
    0x7fe0,0x7fe0,0x3fc0,0x0f00
};


short RequestDialog(const char* request);

/*---------------------------------------------------------------------------*/
/* includes menu code and code of engine                                     */
/*---------------------------------------------------------------------------*/

#include "menu.c"    // menu handling
#include "c4algo.c"  // the AI engine itself
#include "logo.c"    // logo treatment


/*===========================================================================*/
/* clear top of screen                                                       */
/*===========================================================================*/
void ClearTopOfScreen(void) {
    memset(GetPlane(0),0,30*m_offsety);
    memset(GetPlane(1),0,30*m_offsety);
}


/*===========================================================================*/
/* draw coordinates                                                          */
/*===========================================================================*/
void DrawCoordinates(short draw_not) {
    char       s[2];
    short      loop,i;

    s[1] = 0;

    FontSetSys(F_4x6);

    for (loop=0;loop<2;loop++) {
        SetPlane(loop);
        memset(GetPlane(loop),0,30*m_offsety);
        for (i=0;i<m_width;i++) {
            if (i==draw_not) continue;
            if (i>8) s[0] = 'A'+i-9;
            else     s[0] = '1'+i;
            DrawStr(m_offsetx+i*13+5,m_offsety-9,s,A_REPLACE);
        }
    }
}


/*===========================================================================*/
/* moves down a tile with given color from (xpos/ystart) to (xpos/yend)      */
/*===========================================================================*/
void MoveDown(short color,short xpos,short ystart,short yend) {
    unsigned short *data1,*data2;

    //-----------------------------------------------
    // set pointers to sprite data depending on color
    //-----------------------------------------------
    if (color) {
        data1=black1;
        data2=black2;
    }
    else {
        data1=white1;
        data2=white2;
    }

    //-----------------------------------------------
    // draw sprite into initial position
    //-----------------------------------------------
    Sprite16(xpos,ystart,12,data1,plane1,SPRT_OR);
    Sprite16(xpos,ystart,12,data2,plane2,SPRT_OR);

    ystart++;

    //-----------------------------------------------
    // loop until we reach end position
    //-----------------------------------------------
    while (ystart <= yend) {
        WaitForMillis(8);
        //--------------------------
        // clear previously position
        //--------------------------
        Sprite16(xpos,ystart-1,13,cleaner,plane1,SPRT_AND);
        Sprite16(xpos,ystart-1,13,cleaner,plane2,SPRT_AND);
        //----------------------------
        // draw sprite in new position
        //----------------------------
        Sprite16(xpos,ystart,12,data1,plane1,SPRT_OR);
        Sprite16(xpos,ystart,12,data2,plane2,SPRT_OR);
        ystart++;
    }
}


/*===========================================================================*/
/* draw cursor using specified color at top of column i                      */
/*===========================================================================*/
void DrawCursor(short color,short i) {
    unsigned short *data1,*data2;

    //-------------------------------------
    // set pointer to sprite data depending
    // on color
    //-------------------------------------
    if (color) {
        data1=black1;
        data2=black2;
    }
    else {
        data1=white1;
        data2=white2;
    }

    DrawCoordinates(i);
    Sprite16(i*13+1+m_offsetx,m_offsety-13,12,data1,plane1,SPRT_OR);
    Sprite16(i*13+1+m_offsetx,m_offsety-13,12,data2,plane2,SPRT_OR);
}


/*===========================================================================*/
/* draws computer thinking using sprite of specified color                   */
/*===========================================================================*/
void DrawComputerThinking(short color) {
    short           loop;
    unsigned short* data1;
    unsigned short* data2;

    //-------------------------------------
    // set pointer to sprite data depending
    // on color
    //-------------------------------------
    if (color) {
        data1=black1;
        data2=black2;
    }
    else {
        data1=white1;
        data2=white2;
    }


    ClearTopOfScreen();
    //-------------------------------------
    // draw sprite
    //-------------------------------------
    Sprite16(m_think_x,m_think_y,12,data1,plane1,SPRT_OR);
    Sprite16(m_think_x,m_think_y,12,data2,plane2,SPRT_OR);

    //-------------------------------------
    // draw the "Thinking ..." text
    //-------------------------------------
    FontSetSys(F_4x6);
    for (loop=0;loop<2;loop++) {
        SetPlane(loop);
        DrawStr(m_think_x+15,m_think_y+3,m_think_m,A_REPLACE);
    }
}


/*===========================================================================*/
/* initializes the board graphics                                            */
/*===========================================================================*/
static inline void InitScreen(void) {
    short  i;
    short  loop;
    char   s[2];

    s[1] = 0;

    FontSetSys(F_4x6);

    for (loop=0;loop<2;loop++) {
        SetPlane(loop);
        ClearScreen();
        DrawClipRect(MakeWinRect(m_offsetx,
                                 m_offsety,
                                 m_width*13+m_offsetx,
                                 m_height*12+1+m_offsety),
                                 &fullscreen,A_REPLACE);
        //---------------------------------------------------
        // remove top of rectangle by drawing an inverse line
        //---------------------------------------------------
        DrawLine(m_offsetx+1,m_offsety,m_width*13+m_offsetx-1,m_offsety,A_REVERSE);
        if (loop==0) {
            for (i=1;i<m_width;i++) {
                DrawLine(m_offsetx+i*13,
                         m_offsety+1,
                         m_offsetx+i*13,
                         m_height*12+1+m_offsety,A_NORMAL);
            }
        }
    }

    //---------------------------------------------------
    // center the cursor at the middle column
    //---------------------------------------------------
    actual_column = (m_width+1)/2-1;
    DrawCoordinates(-1);
}


/*===========================================================================*/
/* handles cursor movement                                                   */
/* returns 0 if user pressed ESCAPE                                          */
/*         1 otherwise                                                       */
/*===========================================================================*/
static inline short HandleCursor(short player) {
    short input;

    //-----------------------------------------------
    // remove cursor and draw again just to make sure
    // the cursor gets the right color
    //-----------------------------------------------
    DrawCursor(player,actual_column);

    //-----------------------------------------------
    // loop until user drops his piece or presses
    // KEY_ESC
    // if user presses KEY_LEFT or KEY_RIGHT ->
    // move cursor to new position
    //-----------------------------------------------
    do {
        input = GetUserInput();
        if      (input == KEY_ESC) return 0;
        else if (input == KEY_ENTER || input == KEY_DOWN)  return 1;
        else if (input == KEY_LEFT) {
            actual_column--;
            if (actual_column < 0) actual_column = m_width-1;
            DrawCursor(player,actual_column);
        }
        else if (input == KEY_RIGHT) {
            actual_column++;
            if (actual_column >= m_width) actual_column = 0;
            DrawCursor(player,actual_column);
        }
    }
    while (1);
}


/*===========================================================================*/
/* handles a request dialog (returns 1 if user agrees to question)           */
/*===========================================================================*/
short RequestDialog(const char* request) {
    short read;
    short i;
    short sy;
    short loop;

    LCD_BUFFER LCD_tmp1;
    LCD_BUFFER LCD_tmp2;

    memcpy(LCD_tmp1,plane1,LCD_SIZE);
    memcpy(LCD_tmp2,plane2,LCD_SIZE);

    sy = ScrRect->xy.y1/2 - 2;
    for (i=0;i<=10;i++) {
        for (loop=0;loop<2;loop++) {
            SetPlane(loop);
            DrawClipRect(MakeWinRect(0,sy-i,ScrRect->xy.x1-1,sy+i),ScrRect,1);
            DrawClipRect(MakeWinRect(1,sy-i+1,ScrRect->xy.x1-2,sy+i-1),ScrRect,0);
        }
        WaitForMillis(7);
    }

    FontSetSys(F_6x8);
    SetPlane(0);
    DrawStr((ScrRect->xy.x1-DrawStrWidth(request,F_6x8))/2,sy-3,request,A_REPLACE);
    SetPlane(1);
    DrawStr((ScrRect->xy.x1-DrawStrWidth(request,F_6x8))/2,sy-3,request,A_REPLACE);


    do {
        read = GetUserInput();
    }
    while (read != KEY_ENTER && read != KEY_ESC);

    memcpy(plane1,LCD_tmp1,LCD_SIZE);
    memcpy(plane2,LCD_tmp2,LCD_SIZE);

    return read == KEY_ENTER;
}


/*===========================================================================*/
/* play the game (nomen-omen-est)                                            */
/*===========================================================================*/
static inline void PlayGame() {
    short player[2], level[2], turn = 0, move;
    short num_to_connect;
    short y;

    char** board;

    //--------------------------------------------
    // handle parameters which are set by the menu
    // and calculator display globals
    //--------------------------------------------
    m_width  = settings[2].active+5;
    m_height = settings[3].active+5;

    m_offsetx = ((12 - m_width)*13)/2+1+ti92_offsetx;
    m_offsety = (7 - m_height)*6+13+ti92_offsety;

    m_think_l = DrawStrWidth(m_think_m,F_4x6)+15;
    m_think_x = m_offsetx + (m_width*13-m_think_l)/2;
    m_think_y = m_offsety-13;

    num_to_connect = settings[4].active+4;

    if (settings[0].active == 0) {
        player[0] = HUMAN;
    }
    else {
        player[0] = COMPUTER;
        level[0]  = settings[0].active+3;
    }
    if (settings[1].active == 0) {
        player[1] = HUMAN;
    }
    else {
        player[1] = COMPUTER;
        level[1]  = settings[1].active+3;
    }

    //-------------------------
    // initialize board display
    //-------------------------
    InitScreen();

    //-------------------------
    // initialize engine
    //-------------------------
    c4_new_game(m_width, m_height, num_to_connect);

    //-------------------------
    // major game loop ...
    //-------------------------
    do {
        if (kbhit()) {
            if (GetUserInput() == KEY_ESC) {
                if (RequestDialog("Sure to leave Game ?")) {
                    game_in_progress = FALSE;
                    return;
                }
            }
        }
        if (player[turn] == HUMAN) {
            do {
                while (!HandleCursor(turn)) {
                    if (RequestDialog("Sure to leave Game ?")) {
                        game_in_progress = FALSE;
                        return;
                    }
                }
                move = actual_column;
            }
            while (!c4_make_move(turn, move, NULL));


            board = c4_board();
            for (y=m_height-1; y>=0; y--) {
                 if (board[move][y] != C4_NONE) {
                     MoveDown(board[move][y],
                              1+move*13+m_offsetx,
                              1+m_offsety-14,
                              ((m_height-1)-y)*12+1+m_offsety);
                     break;
                 }
            }
        }
        else {
            DrawComputerThinking(turn);
            c4_auto_move(turn, level[turn], &move, NULL);
            ClearTopOfScreen();
            if (break_game) {
                game_in_progress = FALSE;
                return;
            }

            board = c4_board();
            for (y=m_height-1; y>=0; y--) {
                 if (board[move][y] != C4_NONE) {
                     MoveDown(board[move][y],
                              1+move*13+m_offsetx,
                              1+m_offsety-14,
                              ((m_height-1)-y)*12+1+m_offsety);
                     break;
                 }
            }
        }

        if (turn) turn = 0;
        else      turn = 1;

    } while (!c4_is_winner(0) && !c4_is_winner(1) && !c4_is_tie());

    if (!c4_is_winner(0) && !c4_is_winner(1)) RequestDialog("It's a Draw!");
    else {
        //-------------------------------------------------------
        // there exists a winning combination -> let it blink !!!
        //-------------------------------------------------------
        short    winx1,winy1,winx2,winy2;
        short    tempx,tempy;
        short    loop;
        short    key = -1;
        short    stepx;
        short    stepy;
        SCR_RECT rect;

        c4_win_coords(&winx1,&winy1,&winx2,&winy2);

        if (winx1<winx2)      stepx = 1;
        else if (winx1>winx2) stepx = -1;
        else                  stepx = 0;


        if (winy1<winy2)      stepy = 1;
        else if (winy1>winy2) stepy = -1;
        else                  stepy = 0;

        //-----------------------------------------
        // start APD_TIMER again for blinking phase
        //-----------------------------------------
        OSTimerRestart(APD_TIMER);

        //-------------------------------------
        // draw the winning text
        //-------------------------------------
        if (c4_is_winner(0)) {
            char *s = m_think_m;
            m_think_m = (char*)"White wins!";
            DrawComputerThinking(0);
            m_think_m = s;
        }
        else {
            char *s = m_think_m;
            m_think_m = (char*)"Black wins!";
            DrawComputerThinking(1);
            m_think_m = s;
        }

        //-------------------------
        // blinking loop
        //-------------------------
        do {
            tempx = winx1;
            tempy = winy1;
            for (loop=0;loop<num_to_connect;loop++) {
                rect.xy.x0 = 1+tempx*13+m_offsetx;
                rect.xy.y0 = ((m_height-1)-tempy)*12+1+m_offsety;
                rect.xy.x1 = (tempx+1)*13+m_offsetx;
                rect.xy.y1 = ((m_height-1)-tempy+1)*12+m_offsety;
                SetPlane(0);
                ScrRectFill(&rect,&fullscreen,A_XOR);
                SetPlane(1);
                ScrRectFill(&rect,&fullscreen,A_XOR);
                tempx+=stepx;
                tempy+=stepy;
            }
            WaitForMillis(250);
            if (kbhit()) key = GetUserInput();
            else if (OSTimerExpired(APD_TIMER)) {
                off();
                OSTimerRestart(APD_TIMER);
            }
        }
        while (key != KEY_ENTER && key != KEY_ESC);
    }

    game_in_progress = FALSE;
}


/*===========================================================================*/
/* main routine: where all the fun starts ...                                */
/*===========================================================================*/
void _main(void) {
/*
    if (TI89) {
        ti92_offsetx = 0;
        ti92_offsety = 0;
    }
    else {
        ti92_offsetx = 40;
        ti92_offsety = 14;
    }
*/
    //-------------------------------------
    // turn grayscales on
    //-------------------------------------
    if (!GrayOn()) return;

    //-------------------------------------
    // output logo and wait a little bit
    //-------------------------------------
    Logo();

    plane1 = GetPlane(0);
    plane2 = GetPlane(1);

    memset(plane1,0,LCD_SIZE);
    memset(plane2,0,LCD_SIZE);

    //-------------------------------------
    // loop until player resigns ...
    //-------------------------------------
    if (AllocGameMemory()) {
        while (HandleMenu()) PlayGame();
    }
    else {
        RequestDialog("ERROR: NOT ENOUGH MEM");
    }

    //---------------------------------------------
    // cleanup and bye =8-)
    //---------------------------------------------
    FreeGameMemory();
    GrayOff();
}

